import java.util.List;
import java.util.ArrayList;
import java.util.Vector;
import java.util.concurrent.*;
class Executor extends Thread {
	
	private CyclicBarrier counter;
	
	// ------------------------------------------------------------------------
	
	private void setCounter( int nodes ) {
		if ( counter != null ) {
			throw new RuntimeException( "Nested counter" );
		}
		
		counter = new CyclicBarrier( nodes + 1 );
	}

	// ------------------------------------------------------------------------
	
	private void syncThreads() {
		try {
			counter.await();
			counter = null;
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (BrokenBarrierException e) {
			e.printStackTrace();
		}
	}
	
	// ------------------------------------------------------------------------
	
	public synchronized void run() {

		// ==========================================================================================================================================
		// CREATE THE TREE AUTOMATICALLY SO THAT IT CONTAINS 2^k TERMINAL NODES (leaves)
		
		int k = 5;
		int levels_count = k;
		int MAX_TIME_STEPS = 1;
		
		Vector<P1> p1_vect = new Vector<P1>();
		Vector<P2> p2_vect = new Vector<P2>();
		
		int last_production = 0;
		int last_v1_pos = 0;
		int last_v2_pos = 0;
		int last_v2_pos_previous = 0;
		
		int i;
		int j;
		int temp;
		
		// ==========================================================================================================================================
		// LOOP GROUP 1: TREE GENERATION
		
		// Create the root node
	  	
		Vertex S = new Vertex( null, null, null, "S" );
		
		// Create the rest of the tree
		
		for ( i = 1; i <= levels_count; i++ )
		{
			if ( last_production == 0 )
			{
				// No production type has been applied so far, select P1 and the root vertex
				
				setCounter( 1 );
				
				P1 p1 = new P1( S, counter );
				p1.start();
				
				syncThreads();
				
				last_v1_pos = 0;
				p1_vect.add( p1 );
				
				last_production = 1;
			}
			else if ( last_production == 1 )
			{
				// The last production type was P1, now apply P2
				
				last_v2_pos = p2_vect.size();
				temp = p1_vect.size() - last_v1_pos;
				
				for ( j = 0; j < temp; j++ )
				{
					setCounter( 2 );
					
					P2 p2_on_left = new P2( p1_vect.elementAt( last_v1_pos + j ).m_vertex.m_left, counter );
					P2 p2_on_right = new P2( p1_vect.elementAt( last_v1_pos + j ).m_vertex.m_right, counter );
					
					p2_on_left.start();
					p2_on_right.start();
					
					syncThreads();
					
					p2_vect.add( p2_on_left );
					p2_vect.add( p2_on_right );
				}
				
				last_production = 2;
			}
			else
			{
				// The last production type was P2, now apply P2
				
				last_v2_pos_previous = last_v2_pos;
				last_v2_pos = p2_vect.size();
				temp = p2_vect.size() - last_v2_pos_previous;
				
				for ( j = 0; j < temp; j++ )
				{
					setCounter( 2 );
					
					P2 p2_on_left = new P2( p2_vect.elementAt( last_v2_pos_previous + j ).m_vertex.m_left, counter );
					P2 p2_on_right = new P2( p2_vect.elementAt( last_v2_pos_previous + j ).m_vertex.m_right, counter );
					
					p2_on_left.start();
					p2_on_right.start();
					
					syncThreads();
					
					p2_vect.add( p2_on_left );
					p2_vect.add( p2_on_right );
				}
				
				last_production = 2;
			}
		}
		
		// ==========================================================================================================================================
		
		for ( int t = 0; t < MAX_TIME_STEPS; t++ )
		{
			// ----------------------------------------------------------------------------------------------
			// LOOP GROUP 2: MATRICES
	
			//
			// A1--A...A--AN
			//
			
			int inner_loop_range = ( int )Math.pow( 2, levels_count - 1 );
			setCounter( inner_loop_range * 2 );		// Each parent creates 2 children
			
			A1 local_matrix_A1 = null;
			AN local_matrix_AN = null;
			Vector<A> local_matrix_A_vect = new Vector<A>();
			
			double dim_a2 = 0.5;
			
			for ( j = 0; j < inner_loop_range; j++ )
			{
				if ( j == 0 )
				{
					// Left
					
					System.out.println( "A1 (1.0, " + dim_a2 + ", 1.0, " + dim_a2 + ")" );
					local_matrix_A1 = new A1( p2_vect.elementAt( last_v2_pos ).m_vertex.m_left, counter, 1.0, dim_a2, 1.0, dim_a2 );
					
					dim_a2 /= 2.0;
					
					// Right
					
					System.out.println( "A (1.0, " + dim_a2 + ", 1.0, " + dim_a2 + ")" );
					A local_matrix_A = new A( p2_vect.elementAt( last_v2_pos ).m_vertex.m_right, counter, 1.0, dim_a2, 1.0, dim_a2 );
					local_matrix_A_vect.add( local_matrix_A );
					
					dim_a2 /= 2.0;
				}
				else if ( j == inner_loop_range - 1 )
				{
					// Left
					
					System.out.println( "A (1.0, " + dim_a2 + ", 1.0, " + dim_a2 + ")" );
					A local_matrix_A = new A( p2_vect.elementAt( last_v2_pos + j ).m_vertex.m_left, counter, 1.0, dim_a2, 1.0, dim_a2 );
					local_matrix_A_vect.add( local_matrix_A );
					
					// Right
					
					System.out.println( "AN (1.0, " + dim_a2 + ", 1.0, " + dim_a2 + ")" );
					local_matrix_AN = new AN( p2_vect.elementAt( last_v2_pos + j ).m_vertex.m_right, counter, 1.0, dim_a2, 1.0, dim_a2 );
				}
				else
				{
					// Left
					
					System.out.println( "A (1.0, " + dim_a2 + ", 1.0, " + dim_a2 + ")" );
					A local_matrix_A = new A( p2_vect.elementAt( last_v2_pos + j ).m_vertex.m_left, counter, 1.0, dim_a2, 1.0, dim_a2 );
					local_matrix_A_vect.add( local_matrix_A );
					
					dim_a2 /= 2.0;
					
					// Right
					
					System.out.println( "A (1.0, " + dim_a2 + ", 1.0, " + dim_a2 + ")" );
					local_matrix_A = new A( p2_vect.elementAt( last_v2_pos + j ).m_vertex.m_right, counter, 1.0, dim_a2, 1.0, dim_a2 );
					local_matrix_A_vect.add( local_matrix_A );
					
					dim_a2 /= 2.0;
				}
			}
			
			local_matrix_A1.start();
			temp = local_matrix_A_vect.size();
			
			for ( j = 0; j < temp; j++ )
			{
				local_matrix_A_vect.elementAt( j ).start();
			}
			
			local_matrix_AN.start();
			
			syncThreads();
			System.out.println();
			
			// ---- DEBUG ----
			
			for ( j = 0; j < inner_loop_range; j++ )
			{
				System.out.println( "[A1/A/AN] matrix A at leaf " + ( j * 2 + 1 ) );
				MatrixActions.printMatrix( p2_vect.elementAt( last_v2_pos + j ).m_vertex.m_left.m_a );
				
				System.out.println( "[A1/A/AN] matrix A at leaf " + ( j * 2 + 2 ) );
				MatrixActions.printMatrix( p2_vect.elementAt( last_v2_pos + j ).m_vertex.m_right.m_a );
			}
			
			System.out.println();
			
			//
			// A2/E2 .. A2/E2
			// The lowest level above leaves in the tree
			//
			
			Vector<A2> merged_matrix_A2_vect = new Vector<A2>();
			Vector<E2> elimination_E2_vect = new Vector<E2>();
			temp = p2_vect.size();

			for ( i = levels_count - 1; i > 0; i-- )
			{
				inner_loop_range = ( int )Math.pow( 2, i );
				temp -= inner_loop_range;
				
				// --- Apply A2
				
				setCounter( inner_loop_range );
				
				for ( j = 0; j < inner_loop_range; j++ )
				{
					A2 merged_matrix = new A2( p2_vect.elementAt( temp + j ).m_vertex, counter );
					
					merged_matrix.start();
					merged_matrix_A2_vect.add( merged_matrix );
				}
				
				syncThreads();
				
				// --- Apply E2
				
				setCounter( inner_loop_range );
				
				for ( j = 0; j < inner_loop_range; j++ )
				{
					E2 gauss_elimination_matrix = new E2( p2_vect.elementAt( temp + j ).m_vertex, counter );
					
					gauss_elimination_matrix.start();
					elimination_E2_vect.add( gauss_elimination_matrix );
				}
				
				syncThreads();
			}
			
			// ---- DEBUG ----
			
			/*inner_loop_range = ( int )Math.pow( 2, levels_count - 1 );
			temp = p2_vect.size() - inner_loop_range;
			
			for ( j = 0; j < inner_loop_range; j++ )
			{
				System.out.println( "[E2] matrix A at interior node " + ( temp + j + 1 ) );
				MatrixActions.printMatrix( p2_vect.elementAt( temp + j ).m_vertex.m_a );
			}
			
			System.out.println();*/
			
			//
			// Aroot/Eroot
			//
			
			// --- Apply Aroot
			
			setCounter( 1 );
			
			Aroot merged_root_matrix = new Aroot( p1_vect.elementAt( 0 ).m_vertex, counter );
			merged_root_matrix.start();
			
			System.out.println( "[Aroot] matrix A at root" );
			MatrixActions.printMatrix( p1_vect.elementAt( 0 ).m_vertex.m_a );
			System.out.println();
			
			syncThreads();
			
			// --- Apply Eroot
			
			setCounter( 1 );
			
			Eroot full_elimination_matrix = new Eroot( p1_vect.elementAt( 0 ).m_vertex, counter );
			full_elimination_matrix.start();
			
			syncThreads();
			
			// ----------------------------------------------------------------------------------------------
			// LOOP GROUP 3: BACKWARD SUBSTITUTION
			
			// Root
			
			setCounter( 1 );
			
			BS backward_substitution_root = new BS( p1_vect.elementAt( 0 ).m_vertex, counter );
			backward_substitution_root.start();
			
			syncThreads();
			
			// A2
			
			Vector<BS> backward_substitution_vect = new Vector<BS>();
			
			temp = 0;
			int A2_levels_count = levels_count - 1;
			
			for ( i = 1; i <= A2_levels_count; i++ )
			{
				inner_loop_range = ( int )Math.pow( 2, i );
				setCounter( inner_loop_range );
				
				for ( j = 0; j < inner_loop_range; j++ )
				{
					BS backward_substitution = new BS( p2_vect.elementAt( temp + j ).m_vertex, counter );
					backward_substitution.start();
					
					backward_substitution_vect.add( backward_substitution );
				}
				
				syncThreads();
				temp += inner_loop_range;
			}
			
			// ----------------------------------------------------------------------------------------------
			// LOOP GROUP 4: PRINT THE RESULT AND DRAW THE PLOT
			
			System.out.println();
			
			if ( t % 100 == 0 )
			{
				System.out.println( "t = " + t );
				
				/*List< Double > result = new ArrayList< Double >();
				temp = p2_vect.size() - last_v2_pos;
				
				for ( j = 0; j < temp; j++ )
				{
					result.add( p2_vect.elementAt( last_v2_pos + j ).m_vertex.m_x[ 1 ] );
				}
				
				result.add( p2_vect.elementAt( last_v2_pos + j - 1 ).m_vertex.m_x[ 2 ] );
				PlotDrawer.draw( result );
				
				try {
					TimeUnit.MILLISECONDS.sleep( 100 );
				} catch ( InterruptedException e ) {
					e.printStackTrace();
				}*/
			}
		}
		
		System.out.println( "Finished" );
	}
}