import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

abstract class Production extends Thread {

	//vertex where the production will be applied
	Vertex m_vertex;
	
	//graph drawer
	GraphDrawer m_drawer;
	
	//productions counter
	CyclicBarrier m_counter;
	
	public static double dt = 0.001;
	public static double alpha = 0.0;
	public static double beta = 1 - alpha;
	
	// ----------------------------------------------------------------------------
	
	Production( Vertex Vert, CyclicBarrier Count ){
		m_vertex = Vert;
		m_counter = Count;
		m_drawer = new GraphDrawer();
	}
	
	// ----------------------------------------------------------------------------
	//returns first vertex from the left

	abstract Vertex apply(Vertex v);
	
	// ----------------------------------------------------------------------------
	//run the thread
	
	public void run() {
		//apply the production
		m_vertex = apply(m_vertex);
		
		//plot the graph
		m_drawer.draw(m_vertex);
		
		try {
			m_counter.await();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (BrokenBarrierException e) {
			e.printStackTrace();
		}
	}
	
	// ----------------------------------------------------------------------------
}

//----------------------------------------------------------------------------

class ProductionElement extends Production {
	double m_a1;
	double m_a2;
	double m_b1;
	double m_b2;
	
	ProductionElement( Vertex Vert, CyclicBarrier Count, double A1, double A2, double B1, double B2 ) {
		super( Vert, Count );
		
		m_a1 = A1;
		m_a2 = A2;
		m_b1 = B1;
		m_b2 = B2;
	}
	
	@Override
	Vertex apply(Vertex v) {
		return null;
	}
}

// ----------------------------------------------------------------------------
// Production: root node

class P1 extends Production {
	P1( Vertex Vert, CyclicBarrier Count ){
		super( Vert, Count );
	}
	
	Vertex apply(Vertex S) {
	  //System.out.println("p1");
	  
	  Vertex T1 = new Vertex(null,null,S,"T");
	  Vertex T2 = new Vertex(null,null,S,"T");
	  
	  S.set_left(T1);
	  S.set_right(T2);
	  S.set_label("root");
	  
	  return S;
	}
}

//----------------------------------------------------------------------------
// Production: intermediate tree nodes

class P2 extends Production {
	P2( Vertex Vert, CyclicBarrier Count ){
		super( Vert, Count );
	}
	
	Vertex apply(Vertex T) {
	  //System.out.println("p2");
	  
	  Vertex T1 = new Vertex(null,null,T,"T");
	  Vertex T2 = new Vertex(null,null,T,"T");
	  
	  T.set_left(T1);
	  T.set_right(T2);
	  T.set_label("int");
	 
	  return T;
	}
}

//----------------------------------------------------------------------------

class A extends ProductionElement {
	A( Vertex Vert, CyclicBarrier Count, double A1, double A2, double B1, double B2 ){
		super( Vert, Count, A1, A2, B1, B2 );
	}
	
	@Override
	Vertex apply(Vertex T) {
		//System.out.println("A");
		
		//T.m_a[ 1 ][ 1 ] = 1.0 / 9.0 * m_a1 * m_a2;
		T.m_a[ 1 ][ 1 ] = 1.0 / ( 9.0 * m_a1 * m_a2 );
		T.m_a[ 2 ][ 2 ] = T.m_a[ 1 ][ 1 ];
		T.m_a[ 3 ][ 3 ] = T.m_a[ 1 ][ 1 ];
		T.m_a[ 4 ][ 4 ] = T.m_a[ 1 ][ 1 ];
		
		//T.m_a[ 1 ][ 2 ] = -1.0 / 18.0 * m_a1 * m_a2;
		T.m_a[ 1 ][ 2 ] = -1.0 / ( 18.0 * m_a1 * m_a2 );
		T.m_a[ 2 ][ 1 ] = T.m_a[ 1 ][ 2 ];
		
		//T.m_a[ 1 ][ 3 ] = 1.0 / 36.0 * m_a1 * m_a2;
		T.m_a[ 1 ][ 3 ] = 1.0 / ( 36.0 * m_a1 * m_a2 );
		T.m_a[ 3 ][ 1 ] = T.m_a[ 1 ][ 3 ];
		
		//T.m_a[ 1 ][ 4 ] = -1.0 / 18.0 * m_a1 * m_a2;
		T.m_a[ 1 ][ 4 ] = -1.0 / ( 18.0 * m_a1 * m_a2 );
		T.m_a[ 4 ][ 1 ] = T.m_a[ 1 ][ 4 ];
		T.m_a[ 2 ][ 3 ] = T.m_a[ 1 ][ 4 ];
		T.m_a[ 3 ][ 2 ] = T.m_a[ 1 ][ 4 ];
		
		//T.m_a[ 2 ][ 4 ] = 1.0 / 36.0 * m_a1 * m_a2;
		T.m_a[ 2 ][ 4 ] = 1.0 / ( 36.0 * m_a1 * m_a2 );
		T.m_a[ 4 ][ 2 ] = T.m_a[ 2 ][ 4 ];
		
		//T.m_a[ 3 ][ 4 ] = -1.0 / 18.0 * m_a1 * m_a2;
		T.m_a[ 3 ][ 4 ] = -1.0 / ( 18.0 * m_a1 * m_a2 );
		T.m_a[ 4 ][ 3 ] = T.m_a[ 3 ][ 4 ];
		
		T.m_b[ 1 ] = Math.tan( Math.PI * m_b1 + m_a1 / 2.0 );
		T.m_b[ 2 ] = T.m_b[ 1 ];
		T.m_b[ 3 ] = T.m_b[ 1 ];
		T.m_b[ 4 ] = T.m_b[ 1 ];
		
		// Shift to right bottom corner
		
		T.m_a[ 6 ][ 3 ] = T.m_a[ 4 ][ 1 ];
		T.m_a[ 6 ][ 4 ] = T.m_a[ 4 ][ 2 ];
		T.m_a[ 6 ][ 5 ] = T.m_a[ 4 ][ 3 ];
		T.m_a[ 6 ][ 6 ] = T.m_a[ 4 ][ 4 ];
		
		T.m_a[ 5 ][ 3 ] = T.m_a[ 3 ][ 1 ];
		T.m_a[ 5 ][ 4 ] = T.m_a[ 3 ][ 2 ];
		T.m_a[ 5 ][ 5 ] = T.m_a[ 3 ][ 3 ];
		T.m_a[ 5 ][ 6 ] = T.m_a[ 3 ][ 4 ];
		
		T.m_a[ 4 ][ 3 ] = T.m_a[ 2 ][ 1 ];
		T.m_a[ 4 ][ 4 ] = T.m_a[ 2 ][ 2 ];
		T.m_a[ 4 ][ 5 ] = T.m_a[ 2 ][ 3 ];
		T.m_a[ 4 ][ 6 ] = T.m_a[ 2 ][ 4 ];
		
		T.m_a[ 3 ][ 3 ] = T.m_a[ 1 ][ 1 ];
		T.m_a[ 3 ][ 4 ] = T.m_a[ 1 ][ 2 ];
		T.m_a[ 3 ][ 5 ] = T.m_a[ 1 ][ 3 ];
		T.m_a[ 3 ][ 6 ] = T.m_a[ 1 ][ 4 ];
		
		// Reset of unused entries
		
		T.m_a[ 1 ][ 1 ] = 0.0;
		T.m_a[ 1 ][ 2 ] = 0.0;
		T.m_a[ 1 ][ 3 ] = 0.0;
		T.m_a[ 1 ][ 4 ] = 0.0;
		T.m_a[ 1 ][ 5 ] = 0.0;
		T.m_a[ 1 ][ 6 ] = 0.0;
		
		T.m_a[ 2 ][ 1 ] = 0.0;
		T.m_a[ 2 ][ 2 ] = 0.0;
		T.m_a[ 2 ][ 3 ] = 0.0;
		T.m_a[ 2 ][ 4 ] = 0.0;
		T.m_a[ 2 ][ 5 ] = 0.0;
		T.m_a[ 2 ][ 6 ] = 0.0;
		
		T.m_a[ 3 ][ 1 ] = 0.0;
		T.m_a[ 3 ][ 2 ] = 0.0;
		
		T.m_a[ 4 ][ 1 ] = 0.0;
		T.m_a[ 4 ][ 2 ] = 0.0;
		
		T.m_a[ 5 ][ 1 ] = 0.0;
		T.m_a[ 5 ][ 2 ] = 0.0;
		
		T.m_a[ 6 ][ 1 ] = 0.0;
		T.m_a[ 6 ][ 2 ] = 0.0;
		
		return T;
	}
}

//----------------------------------------------------------------------------

class A1 extends A {
	A1( Vertex Vert, CyclicBarrier Count, double A1, double A2, double B1, double B2 ){
		super( Vert, Count, A1, A2, B1, B2 );
	}
}

//----------------------------------------------------------------------------

class AN extends A {
	AN( Vertex Vert, CyclicBarrier Count, double A1, double A2, double B1, double B2 ){
		super( Vert, Count, A1, A2, B1, B2 );
	}
}

//----------------------------------------------------------------------------
//Merging of frontal matrices and eliminating fully assembled rows

class A2 extends Production {
	A2( Vertex Vert, CyclicBarrier Count ){
		super( Vert, Count );
	}
	
	Vertex apply(Vertex T) {
		//System.out.println("A2");
		
		// Merging 4x4 into 6x6 (A1/A/AN lub E2) on the diagonal, get new 6x6
		// Final matrix for elimination:
		
		T.m_a[3][1] = T.m_left.m_a[3][5];
		T.m_a[3][2] = T.m_left.m_a[3][6];
		T.m_a[3][3] = T.m_left.m_a[3][3];
		T.m_a[3][4] = T.m_left.m_a[3][4];
		
		T.m_a[4][1] = T.m_left.m_a[4][5];
		T.m_a[4][2] = T.m_left.m_a[4][6];
		T.m_a[4][3] = T.m_left.m_a[4][3];
		T.m_a[4][4] = T.m_left.m_a[4][4];
		
		T.m_a[1][1] = T.m_left.m_a[5][5] + T.m_right.m_a[3][3];
		T.m_a[1][2] = T.m_left.m_a[5][6] + T.m_right.m_a[3][4];
		T.m_a[1][3] = T.m_left.m_a[5][3];
		T.m_a[1][4] = T.m_left.m_a[5][4];
		
		T.m_a[2][1] = T.m_left.m_a[6][5] + T.m_right.m_a[4][3];
		T.m_a[2][2] = T.m_left.m_a[6][6] + T.m_right.m_a[4][4];
		T.m_a[2][3] = T.m_left.m_a[6][3];
		T.m_a[2][4] = T.m_left.m_a[6][4];
		
		T.m_a[5][1] = T.m_right.m_a[5][3];
		T.m_a[5][2] = T.m_right.m_a[5][4];
		T.m_a[5][5] = T.m_right.m_a[5][5];
		T.m_a[5][6] = T.m_right.m_a[5][6];
		
		T.m_a[6][1] = T.m_right.m_a[6][3];
		T.m_a[6][2] = T.m_right.m_a[6][4];
		T.m_a[6][5] = T.m_right.m_a[6][5];
		T.m_a[6][6] = T.m_right.m_a[6][6];
		
		T.m_b[1] = T.m_left.m_b[3] + T.m_right.m_b[1];
		T.m_b[2] = T.m_left.m_b[4] + T.m_right.m_b[2];
		T.m_b[3] = T.m_left.m_b[1];
		T.m_b[4] = T.m_left.m_b[2];
		T.m_b[5] = T.m_right.m_b[2];
		T.m_b[6] = T.m_right.m_b[3];
		
		return T;
	}
}

//----------------------------------------------------------------------------

class PFEProduction extends Production {
	PFEProduction( Vertex Vert, CyclicBarrier Count ){
		super( Vert, Count );
	}
	
	//----------------------------
	
	Vertex partial_forward_elimination( Vertex T, int elim, int size ) {
		for ( int irow = 1; irow <= elim; irow++ ) {
			double diag = T.m_a[ irow ][ irow ];
			
			for ( int icol = irow; icol <= size; icol++ ) {
				T.m_a[ irow ][ icol ] /= diag;
			}
			
			for ( int isub = irow + 1; isub <= size; isub++ ) {
				double mult = T.m_a[ isub ][ irow ];
				
				for ( int icol = irow; icol <= size; icol++ ) {
					T.m_a[ isub ][ icol ] -= T.m_a[ irow ][ icol ] * mult;
				}
			}
		}
		
		return T;
	}
	
	//----------------------------
	
	Vertex partial_backward_substitution( Vertex T, int elim, int size ) {
		for ( int irow = elim; irow >= 1; irow-- ) {
			T.m_x[ irow ] = 0.0;
			
			for ( int icol = irow + 1; icol <= size; icol++ ) {
				T.m_x[ irow ] += T.m_a[ irow ][ icol ] * T.m_x[ icol ];
			}
			
			T.m_x[ irow ] /= T.m_a[ irow ][ irow ];
		}
		
		return T;
	}

	//----------------------------
	
	@Override
	Vertex apply(Vertex v) {
		return null;
	}
}

//----------------------------------------------------------------------------

class E2 extends PFEProduction {
	E2( Vertex Vert, CyclicBarrier Count ){
		super( Vert, Count );
	}
	
	Vertex apply(Vertex T) {
		//System.out.println("E2");

                T = partial_forward_elimination( T, 2, 6 );

		
		return T;
	}
}

//----------------------------------------------------------------------------

class Aroot extends A2 {
	Aroot( Vertex Vert, CyclicBarrier Count ){
		super( Vert, Count );
	}
}

//----------------------------------------------------------------------------

class Eroot extends PFEProduction {
	Eroot( Vertex Vert, CyclicBarrier Count ){
		super( Vert, Count );
	}
	
	Vertex apply(Vertex T) {
		//System.out.println("Eroot");
		
		T = partial_forward_elimination( T, 6, 6 );
		
		System.out.println( "[Eroot] Matrix A at root, after the elimination" );
		MatrixActions.printMatrix( T.m_a );
		
		T = partial_backward_substitution( T, 6, 6 );
		
		return T;
	}
}

//----------------------------------------------------------------------------
// Productions for backward substitution

class BS extends PFEProduction {
	BS( Vertex Vert, CyclicBarrier Count ){
		super( Vert, Count );
	}
	
	@Override
	Vertex apply(Vertex T) {
		//System.out.println("BS");
		
		T = partial_backward_substitution( T, 2, 6 );
		
		return T;
	}
}

//----------------------------------------------------------------------------
